//---------------------------------
// 2019/6/30

#include <SPI.h>
#include "BME280SPI.h"



// Read length Byte from address by SPI
void BME280SPI::readAddress(uint8_t addr, uint8_t*data, uint32_t length){
  SPI.begin();
  //SPI.setDataMode(SPI_MODE2);
  //SPI.setBitOrder(MSBFIRST);
  //SPI.setClockDivider(SPI_CLOCK_DIV8);
  pinMode(pinCS, OUTPUT);
  digitalWrite(pinCS, LOW);
  SPI.transfer(addr);
  for(uint8_t i=0; i<length; i++){
    data[i] = SPI.transfer(0x0);
  }
  digitalWrite(pinCS, HIGH);
  SPI.end();
  return;
}

// Write 1Byte at specific address by SPI
void BME280SPI::writeAddress(uint8_t addr, uint8_t data){
  SPI.begin();
  //SPI.setDataMode(SPI_MODE2);
  //SPI.setBitOrder(MSBFIRST);
  //SPI.setClockDivider(SPI_CLOCK_DIV8);
  pinMode(pinCS, OUTPUT);
  digitalWrite(pinCS, LOW);
  SPI.transfer(addr & 0x7F);
  SPI.transfer(data);
  digitalWrite(pinCS, HIGH);
  SPI.end();
  return;
}

// ID Read by SPI
uint8_t BME280SPI::idRead(){
  uint8_t data[1];
  readAddress(0xD0, data, 1);
  if(data[0]!=0x60){
    return 1;
  }
  return 0;
}

// Status read
uint8_t BME280SPI::statusRead(){
  uint8_t data[1];
  readAddress(0xF3, data, 1);
  return data[0]&0x9;
}

// Read calibration
void BME280SPI::readCalibration(){
  uint8_t buf[2];
  readAddress(0x88, cal.byte, calLengthTP);  // Cal T and P are read sequentially

  // H1
  readAddress(0xA1, buf, 1);
  cal.byte[24] = buf[0];

  // H2
  readAddress(0xE1, buf, 1);
  cal.byte[26] = buf[0];
  readAddress(0xE2, buf, 1);
  cal.byte[27] = buf[0];

  // H3
  readAddress(0xE3, buf, 1);
  cal.byte[28] = buf[0];

  // H4
  readAddress(0xE4, buf, 2);
  cal.byte[30] = ((buf[0]&0xF)<<4)+(buf[1]&0xF);
  cal.byte[31] = buf[0]>>4;

  // H5
  readAddress(0xE5, buf, 2);
  cal.byte[32] = (buf[0]>>4)+((buf[1]&0xF)<<4);
  cal.byte[33] = buf[1]>>4;

  // H6
  readAddress(0xE7, buf, 1);
  cal.byte[34] = buf[0];
}


// Read ADC registers of T, P and H
// Calculated compensate data
void BME280SPI::readMeas(){
  uint8_t buf[8];
  readAddress(0xF7, buf, 8);
  adcP = (buf[0]<<12) + (buf[1]<<4) + (buf[2]>>4);
  adcT = (buf[3]<<12) + (buf[4]<<4) + (buf[5]>>4);
  adcH = (buf[6]<<8) + buf[7];
  cT = compensate_T_int32(adcT);
  cP = compensate_P_int64(adcP);
  cH = compensate_H_int32(adcH);
  T = ((float)(int32_t)cT/100);
  P = ((float)cP)/25600;
  H = ((float)cH)/1024;
}

// Measure forced mode
void BME280SPI::forced(){
  writeAddress(0xF5, 0x0);  // config
  writeAddress(0xF2, 0x5);  // ctrl_hum, oversampling x16
  writeAddress(0xF4, 0xB5);  // ctrl_meas, oversampling x16, forced mode
  delay(1);
  while(0!=statusRead()){
    delay(1);
  }
}


// Returns temperature in DegC, resolution is 0.01 DegC. Output value of "5123" equals 51.23 DegC.
// t_fine carries fine temperature as global value
BME280_S32_t BME280SPI::compensate_T_int32(BME280_S32_t adc_T)
{
  BME280_S32_t var1, var2, T;
  var1  = ((((adc_T>>3) - ((BME280_S32_t)cal.dig_T1<<1))) * ((BME280_S32_t)cal.dig_T2)) >> 11;
  var2  = (((((adc_T>>4) - ((BME280_S32_t)cal.dig_T1)) * ((adc_T>>4) - ((BME280_S32_t)cal.dig_T1))) >> 12) *
    ((BME280_S32_t)cal.dig_T3)) >> 14;
  t_fine = var1 + var2;
  T  = (t_fine * 5 + 128) >> 8;
  return T;
}

// Returns pressure in Pa as unsigned 32 bit integer in Q24.8 format (24 integer bits and 8 fractional bits).
// Output value of "24674867" represents 24674867/256 = 96386.2 Pa = 963.862 hPa
BME280_U32_t BME280SPI::compensate_P_int64(BME280_S32_t adc_P)
{
  BME280_S64_t var1, var2, p;
  var1 = ((BME280_S64_t)t_fine) - 128000;
  var2 = var1 * var1 * (BME280_S64_t)cal.dig_P6;
  var2 = var2 + ((var1*(BME280_S64_t)cal.dig_P5)<<17);
  var2 = var2 + (((BME280_S64_t)cal.dig_P4)<<35);
  var1 = ((var1 * var1 * (BME280_S64_t)cal.dig_P3)>>8) + ((var1 * (BME280_S64_t)cal.dig_P2)<<12);
  var1 = (((((BME280_S64_t)1)<<47)+var1))*((BME280_S64_t)cal.dig_P1)>>33;
  if (var1 == 0)
  {
    return 0; // avoid exception caused by division by zero
  }
  p = 1048576-adc_P;
  p = (((p<<31)-var2)*3125)/var1;
  var1 = (((BME280_S64_t)cal.dig_P9) * (p>>13) * (p>>13)) >> 25;
  var2 = (((BME280_S64_t)cal.dig_P8) * p) >> 19;
  p = ((p + var1 + var2) >> 8) + (((BME280_S64_t)cal.dig_P7)<<4);
  return (BME280_U32_t)p;
}

// Returns humidity in %RH as unsigned 32 bit integer in Q22.10 format (22 integer and 10 fractional bits).
// Output value of "47445" represents 47445/1024 = 46.333 %RH
BME280_U32_t BME280SPI::compensate_H_int32(BME280_S32_t adc_H)
{
  BME280_S32_t v_x1_u32r;
  v_x1_u32r = (t_fine - ((BME280_S32_t)76800));
  v_x1_u32r = (((((adc_H << 14) - (((BME280_S32_t)cal.dig_H4) << 20) - (((BME280_S32_t)cal.dig_H5) * v_x1_u32r)) +
    ((BME280_S32_t)16384)) >> 15) * (((((((v_x1_u32r * ((BME280_S32_t)cal.dig_H6)) >> 10) * (((v_x1_u32r *
    ((BME280_S32_t)cal.dig_H3)) >> 11) + ((BME280_S32_t)32768))) >> 10) + ((BME280_S32_t)2097152)) *
    ((BME280_S32_t)cal.dig_H2) + 8192) >> 14));
  v_x1_u32r = (v_x1_u32r - (((((v_x1_u32r >> 15) * (v_x1_u32r >> 15)) >> 7) * ((BME280_S32_t)cal.dig_H1)) >> 4));
  v_x1_u32r = (v_x1_u32r < 0 ? 0 : v_x1_u32r);
  v_x1_u32r = (v_x1_u32r > 419430400 ? 419430400 : v_x1_u32r);
  return (BME280_U32_t)(v_x1_u32r>>12); 
}


BME280SPI::BME280SPI(uint8_t cs){
  pinCS = cs;
  stat = statNotStarted;
  adcT = 0;
  adcP = 0;
  adcH = 0;
  cT = 0;
  cP = 0;
  cH = 0;
  T = 0;
  P = 0;
  H = 0;

  for(uint8_t i=0; i<calLength; i++){
    cal.byte[i] = 0;
  }
}


// Measure by forced mode
//  Return 0 if success
uint8_t BME280SPI::meas(){
  // ID Read
  if(0!=idRead()){
    stat = statSensorError;
    return stat;
  }

  readCalibration();
  forced();
  readMeas();

  stat = statSuccess;
  return stat;
}


// Return true if measurement is done
bool BME280SPI::getResult(){
  if(stat == statSuccess){
    return true;
  }else{
    return false;
  }
}


// Print calibration
void BME280SPI::printCalibration(){
  char c[32];
  Serial.print("Calibration Data\n");
  sprintf(c, " dig_T1 : %d\n", cal.dig_T1);Serial.print(c);
  sprintf(c, " dig_T2 : %d\n", cal.dig_T2);Serial.print(c);
  sprintf(c, " dig_T3 : %d\n", cal.dig_T3);Serial.print(c);
  sprintf(c, " dig_P1 : %d\n", cal.dig_P1);Serial.print(c);
  sprintf(c, " dig_P2 : %d\n", cal.dig_P2);Serial.print(c);
  sprintf(c, " dig_P3 : %d\n", cal.dig_P3);Serial.print(c);
  sprintf(c, " dig_P4 : %d\n", cal.dig_P4);Serial.print(c);
  sprintf(c, " dig_P5 : %d\n", cal.dig_P5);Serial.print(c);
  sprintf(c, " dig_P6 : %d\n", cal.dig_P6);Serial.print(c);
  sprintf(c, " dig_P7 : %d\n", cal.dig_P7);Serial.print(c);
  sprintf(c, " dig_P8 : %d\n", cal.dig_P8);Serial.print(c);
  sprintf(c, " dig_P9 : %d\n", cal.dig_P9);Serial.print(c);
  sprintf(c, " dig_H1 : %d\n", cal.dig_H1);Serial.print(c);
  sprintf(c, " dig_H2 : %d\n", cal.dig_H2);Serial.print(c);
  sprintf(c, " dig_H3 : %d\n", cal.dig_H3);Serial.print(c);
  sprintf(c, " dig_H4 : %d\n", cal.dig_H4);Serial.print(c);
  sprintf(c, " dig_H5 : %d\n", cal.dig_H5);Serial.print(c);
  sprintf(c, " dig_H6 : %d\n", cal.dig_H6);Serial.print(c);
}

// Print ADC and compensated register
void BME280SPI::printRegister(){
  char c[32];
  Serial.print("Register Data\n");
  sprintf(c, " ADC T : %x\n", adcT);Serial.print(c);
  sprintf(c, " ADC P : %x\n", adcP);Serial.print(c);
  sprintf(c, " ADC H : %x\n", adcH);Serial.print(c);
  sprintf(c, " cT : %d\n", cT);Serial.print(c);
  sprintf(c, " cP : %d\n", cP);Serial.print(c);
  sprintf(c, " cH : %d\n", cH);Serial.print(c);
}

// Print Measurement data
void BME280SPI::printMeas(){
  Serial.print("Measured Data\n");
  Serial.println(String() + " Temp : " + getMeasStr(selectTemp) + "C");
  Serial.println(String() + " Pressure : " + getMeasStr(selectPressure) + "hPa");
  Serial.println(String() + " Humidity : " + getMeasStr(selectHumidity) + "%");
}

// Get T, P or H
float BME280SPI::getMeas(uint8_t select){
  switch(select){
  case selectTemp:
    return T;
  case selectPressure:
    return P;
  case selectHumidity:
    return H;
  default:
    return 0;
  }
}

// Get T, P or H by String
String BME280SPI::getMeasStr(uint8_t select){
  char c[32];
  switch(select){
  case selectTemp:
    dtostrf(T, 3, 1, c);
    return String(c);
  case selectPressure:
    dtostrf(P, 4, 1, c);
    return String(c);
  case selectHumidity:
    dtostrf(H, 3, 1, c);
    return String(c);
  default:
    return String();
  }
}
