#include <ESP8266WiFi.h>
#include "BME280SPI.h"



//---------------------------------
// Constant

const uint8_t pinLED = 14;
const uint8_t pinSW = 0;
const uint8_t pinBME280CS = 15;
const uint8_t pinSDCS = 2;



//---------------------------------
// Variable

BME280SPI bme280(pinBME280CS);
uint8_t led = 0;

//---------------------------------
// Function

void setup() {
  Serial.begin(115200);
  delay(1);
  Serial.print("\n\n-----------------------\n");
  Serial.print("Program Start\n");
  
  pinMode(pinLED, OUTPUT);   // LED
  pinMode(pinSW, INPUT);     // Switch
  digitalWrite(pinLED, led); // Blink

  // SPI CS High
  pinMode(pinBME280CS, OUTPUT);
  pinMode(pinSDCS, OUTPUT);
  digitalWrite(pinBME280CS, HIGH);
  digitalWrite(pinSDCS, HIGH);

  // BME280 Measure until success
  while(BME280SPI::statSuccess!=bme280.meas()){
    Serial.print("BME280 is not available\n");
    delay(5000);
  }

  bme280.printCalibration();  // Print calibration data of the sensor module
  bme280.printRegister();     // Print measured data registers of the sensor module
  bme280.printMeas();         // Print calibrated measured data
  Serial.print("\n");
}


void loop() {
  Serial.print("BME280 Measure\n");

  if(BME280SPI::statSuccess==bme280.meas()){
    // You can get String format measured data
    Serial.print("Temp : " + bme280.getMeasStr(BME280SPI::selectTemp) + "C\n");
    Serial.print("Pressure : " + bme280.getMeasStr(BME280SPI::selectPressure) + "hPa\n");
    Serial.print("Humidity : " + bme280.getMeasStr(BME280SPI::selectHumidity) + "%\n\n");
    
    led^=1;
    pinMode(pinLED, OUTPUT);   // Need to set mode after SPI use
    digitalWrite(pinLED, led); // Blink
  }else{
    Serial.print("BME280 is not available\n\n");
  }
  
  delay(5000);
}




